/*
 * @file <randomWalk.h>
 *
 * @author Miquel Kegeleirs - <mkegelei@ulb.ac.be>
 *
 * @license MIT License
 */
#include "randomWalk.h"
#include <stdio.h>
#include <stdlib.h>
#include <time.h>

using namespace std;

CEPuckRos::CEPuckRos() :
  m_pcWheels(NULL),
  m_pcProximity(NULL),
  velocity(2.5f),
  speedLeft(0.0f),
  speedRight(0.0f),
  min_range(0.002),
  max_range(0.03),
  calibStep(1000),
  calibAngle(4.6),
  xPos(0.0),
  yPos(0.0),
  theta(0.0),
  randomWalk(1),
  mu(3.0),
  gamma(0.0),
  obstacleAvoidance(0),
  lMin(1.0),
  bias(0.0),
  obstacleSensitivity(0.3),
  maxTurnSteps(25) {}


void CEPuckRos::Init(TConfigurationNode& t_node) 
{
  //get actuators and sensors interface
  m_pcWheels      = GetActuator<CCI_EPuckWheelsActuator              >("epuck_wheels"           );
  m_pcProximity   = GetSensor  <CCI_EPuckProximitySensor             >("epuck_proximity"        );

  //init random numbers
  m_pcRng = CRandom::CreateRNG("argos");
  m_cRandomRange.SetMax(1.0);
  srand (time(NULL));

  //init parameters from argos file
  GetNodeAttributeOrDefault(t_node, "velocity", velocity, velocity);
  GetNodeAttributeOrDefault(t_node, "xPos", xPos, xPos);
  GetNodeAttributeOrDefault(t_node, "yPos", yPos, yPos);
  GetNodeAttributeOrDefault(t_node, "theta", theta, theta);
  GetNodeAttributeOrDefault(t_node, "min_range", min_range, min_range);
  GetNodeAttributeOrDefault(t_node, "max_range", max_range, max_range);
  GetNodeAttributeOrDefault(t_node, "calibAngle", calibAngle, calibAngle);
  GetNodeAttributeOrDefault(t_node, "calibStep", calibStep, calibStep);
  GetNodeAttributeOrDefault(t_node, "randomWalk", randomWalk, randomWalk);
  GetNodeAttributeOrDefault(t_node, "mu", mu, mu);
  GetNodeAttributeOrDefault(t_node, "gamma", gamma, gamma);
  GetNodeAttributeOrDefault(t_node, "obstacleAvoidance", obstacleAvoidance, obstacleAvoidance);
  GetNodeAttributeOrDefault(t_node, "lMin", lMin, lMin);
  GetNodeAttributeOrDefault(t_node, "bias", bias, bias);
  GetNodeAttributeOrDefault(t_node, "obstacleSensitivity", obstacleSensitivity, obstacleSensitivity);
  GetNodeAttributeOrDefault(t_node, "maxTurnSteps", maxTurnSteps, maxTurnSteps);
  
  //init obstacles avoidance
  obstacle = false;
  turnCounter = 0;

  //update initial position
  int id = atoi(GetId().substr(2, 3).c_str());
  yPos += 0.09*id;

  //init odometry
  stepLength = 0;
  stepAngle = CRadians::ZERO;

  //init wheels velocity
  speedLeft = velocity;
  speedRight = velocity;

  //init ROS
  initRos();
}

void CEPuckRos::initRos()
{
  //init epuck name
  std::stringstream name;
  name.str("");
  name << GetId();

  //start ros
  if(!ros::isInitialized())
  {
    char** argv = NULL;
    int argc = 0;
    ros::init(argc, argv, name.str());
  }

  //ros access node
  ros::NodeHandle rosNode;

  //init starting position for map merging
  std::stringstream ss;
  ss << name.str() << "/map_merge/init_pose_x";
  rosNode.setParam(ss.str(), xPos);
  ss.str("");
  ss << name.str() << "/map_merge/init_pose_y";
  rosNode.setParam(ss.str(), yPos);
  ss.str("");
  ss << name.str() << "/map_merge/init_pose_z";
  rosNode.setParam(ss.str(), 0.0);
  ss.str("");
  ss << name.str() << "/map_merge/init_pose_yaw";
  rosNode.setParam(ss.str(), theta);
  
  //init proximity 
  for (int i = 0; i < 8; ++i)
  {
    ss.str("");
    ss << name.str() << "/proximity";
    proximity_pubs[i] = rosNode.advertise<sensor_msgs::Range>(ss.str(), 10);
    proximity_msgs[i].radiation_type = sensor_msgs::Range::INFRARED;
    ss.str("");
    ss << name.str() << "/base_prox" << i;
    proximity_msgs[i].header.frame_id = ss.str();
    proximity_msgs[i].field_of_view = 0.26;
    proximity_msgs[i].min_range = min_range;
      proximity_msgs[i].max_range = max_range;
  }

  //init laser scan
  ss.str("");
  ss << name.str() << "/base_laser";
  laserMsg.header.frame_id = ss.str();
  laserMsg.angle_min = -M_PI;
  laserMsg.angle_max = M_PI;
  laserMsg.angle_increment = M_PI/4.0;
  laserMsg.range_min = min_range+ROBOT_RADIUS; // laser scan situated in the center of the robot
  laserMsg.range_max = max_range+ROBOT_RADIUS; // laser scan situated in the center of the robot
  laserMsg.ranges.resize(8);
  laserMsg.intensities.resize(8);
  ss.str("");
  ss << name.str() << "/scan";
  laserPublisher = rosNode.advertise<sensor_msgs::LaserScan>(ss.str(), 10);
 
  //init odometry
  ss.str("");
  ss << name.str() << "/odom";
  odomPublisher = rosNode.advertise<nav_msgs::Odometry>(ss.str(), 10);

  //init clocks
  currentTime = ros::Time::now();
  lastTime = ros::Time::now();

}

void CEPuckRos::ControlStep() 
{
  if(ros::ok())
  {
    //random walk
    walkStep();
    
    //send ROS messages
    rosControlStep();
    ros::spinOnce();
  }
}

void CEPuckRos::walkStep()
{
  if(randomWalk == 1) //if not ballistic motion
  {
    if(stepLength <= 0) //need to select new direction
    {
      Real r = m_pcRng->Uniform(m_cRandomRange);
      stepLength = ceil(lMin*pow(r, 1/(1-mu))+0.5);
      stepAngle = CRadians(2*atan(((1-gamma)/(1+gamma))*tan(M_PI*(r-0.5))) + bias);
    }
    else
    {
      stepLength--;
    }
  }

  if(obstacleAvoidance == 0) //no obstacle avoidance
  {
    CVector2 speeds = ComputeWheelsVelocityFromVector(CVector2(1.0, stepAngle));
    speedLeft = speeds.GetX();
    speedRight = speeds.GetY();
  }
  else if(obstacleAvoidance == 1) //random rotation
  {
    randomTurnAvoid();
  }
  else if(obstacleAvoidance == 2) //vector field
  {
    vectorFieldAvoid();
  }
  else //error
  {
    speedLeft = 0.0;
    speedRight = 0.0;
  }
  m_pcWheels->SetLinearVelocity(speedLeft, speedRight);
  
}

void CEPuckRos::vectorFieldAvoid()
{
  CVector2 speedVec = CVector2(1.0, stepAngle);
  for(int i=0; i<8; i++) 
  { 
    if(m_pcProximity->GetReadings()[i].Value > obstacleSensitivity) 
    {
      speedVec += CVector2(-1.0, m_pcProximity->GetReadings()[i].Angle);
    }    
  }
  speedVec.Normalize();
  CVector2 speeds = ComputeWheelsVelocityFromVector(speedVec);
  speedLeft = speeds.GetX();
  speedRight = speeds.GetY();
}

void CEPuckRos::randomTurnAvoid()
{
  if(!obstacle)
  {
    Real fMaxReadVal = m_pcProximity->GetReadings()[0].Value;
    UInt32 unMaxReadIdx = 0;
    if(fMaxReadVal < m_pcProximity->GetReadings()[1].Value) 
    {
      fMaxReadVal = m_pcProximity->GetReadings()[1].Value;
      unMaxReadIdx = 1;
    }
    if(fMaxReadVal < m_pcProximity->GetReadings()[7].Value) 
    {
      fMaxReadVal = m_pcProximity->GetReadings()[7].Value;
      unMaxReadIdx = 7;
    }
    if(fMaxReadVal < m_pcProximity->GetReadings()[6].Value) 
    {
      fMaxReadVal = m_pcProximity->GetReadings()[6].Value;
      unMaxReadIdx = 6;
    }
    /* Do we have an obstacle in front? */
    if(fMaxReadVal > obstacleSensitivity) 
    {
      obstacle = true;
      turnCounter = rand() % maxTurnSteps + 1;
      /* Yes, we do: avoid it */
      if(unMaxReadIdx == 0 || unMaxReadIdx == 1) 
      {
        /* The obstacle is on the left, turn right */
        speedLeft = velocity;
        speedRight = 0.0;
      }
      else 
      {
        /* The obstacle is on the right, turn left */
        speedLeft = 0.0;
        speedRight = velocity;
      }
    }
    else
    {
      CVector2 speeds = ComputeWheelsVelocityFromVector(CVector2(1.0, stepAngle));
      speedLeft = speeds.GetX();
      speedRight = speeds.GetY();
    }
  }
  else
  {
    turnCounter--;
    if(turnCounter <= 0)
    {
      obstacle = false;
      CVector2 speeds = ComputeWheelsVelocityFromVector(CVector2(1.0, stepAngle));
      speedLeft = speeds.GetX();
      speedRight = speeds.GetY();
    }
  }
}

void CEPuckRos::rosControlStep()
{
  static tf::TransformBroadcaster br; //needed for tf transform

  std::stringstream name;
  name.str("");
  name << GetId();

  //proximity messages
  for(int i=0; i<8; i++) 
  {
    if(m_pcProximity->GetReadings()[i].Value > 0) 
    {
        proximity_msgs[i].range = (1-sqrt(m_pcProximity->GetReadings()[i].Value))*proximity_msgs[i].max_range;
    } 
    else
    {
        proximity_msgs[i].range = proximity_msgs[i].max_range;
    }
    proximity_msgs[i].header.stamp = ros::Time::now();
    proximity_pubs[i].publish(proximity_msgs[i]);
  }

  //proximity transforms
  std::stringstream parent;
  std::stringstream child;
  tf::Transform transform;
  tf::Quaternion q;

  transform.setOrigin( tf::Vector3(0.035, -0.010, 0.034) );
  q.setRPY(0, 0, 6.11);
  transform.setRotation(q);
  parent << name.str() << "/base_prox0";
  child << name.str() << "/base_link";
  br.sendTransform(tf::StampedTransform(transform, ros::Time::now(), child.str(), parent.str()));

  transform.setOrigin( tf::Vector3(0.025, -0.025, 0.034) );
  q.setRPY(0, 0, 5.59);
  transform.setRotation(q);
  parent.str("");
  parent << name.str() << "/base_prox1";
  br.sendTransform(tf::StampedTransform(transform, ros::Time::now(), child.str(), parent.str()));

  transform.setOrigin( tf::Vector3(0.000, -0.030, 0.034) );
  q.setRPY(0, 0, 4.71);
  transform.setRotation(q);
  parent.str("");
  parent << name.str() << "/base_prox2";
  br.sendTransform(tf::StampedTransform(transform, ros::Time::now(), child.str(), parent.str()));

  transform.setOrigin( tf::Vector3(-0.035, -0.020, 0.034) );
  q.setRPY(0, 0, 3.49);
  transform.setRotation(q);
  parent.str("");
  parent << name.str() << "/base_prox3";
  br.sendTransform(tf::StampedTransform(transform, ros::Time::now(), child.str(), parent.str()));

  transform.setOrigin( tf::Vector3(-0.035, 0.020, 0.034) );
  q.setRPY(0, 0, 2.8);
  transform.setRotation(q);
  parent.str("");
  parent << name.str() << "/base_prox4";
  br.sendTransform(tf::StampedTransform(transform, ros::Time::now(), child.str(), parent.str()));

  transform.setOrigin( tf::Vector3(0.000, 0.030, 0.034) );
  q.setRPY(0, 0, 1.57);
  transform.setRotation(q);
  parent.str("");
  parent << name.str() << "/base_prox5";
  br.sendTransform(tf::StampedTransform(transform, ros::Time::now(), child.str(), parent.str()));

  transform.setOrigin( tf::Vector3(0.025, 0.025, 0.034) );
  q.setRPY(0, 0, 0.70);
  transform.setRotation(q);
  parent.str("");
  parent << name.str() << "/base_prox6";
  br.sendTransform(tf::StampedTransform(transform, ros::Time::now(), child.str(), parent.str()));

  transform.setOrigin( tf::Vector3(0.035, 0.010, 0.034) );
  q.setRPY(0, 0, 0.17);
  transform.setRotation(q);
  parent.str("");
  parent << name.str() << "/base_prox7";
  br.sendTransform(tf::StampedTransform(transform, ros::Time::now(), child.str(), parent.str()));


  //laser scan
  //transform the 8 sensors values to 19 scan values
  float tempProx;
  tempProx = proximity_msgs[4].range;
  if(tempProx >= 0) 
  {
      laserMsg.ranges[0] = tempProx+ROBOT_RADIUS;  
  } 
  else 
  { 
      laserMsg.ranges[0] = laserMsg.range_max;
  }

  tempProx = proximity_msgs[5].range;
  if(tempProx >= 0) 
  {
      laserMsg.ranges[1] = tempProx+ROBOT_RADIUS; 
  } 
  else 
  { 
      laserMsg.ranges[1] = laserMsg.range_max;
  }
  
  tempProx = proximity_msgs[6].range;
  if(tempProx >= 0) 
  {
      laserMsg.ranges[2] = tempProx+ROBOT_RADIUS; 
  } 
  else 
  { 
      laserMsg.ranges[2] = laserMsg.range_max;
  }
  
  tempProx = proximity_msgs[7].range;
  if(tempProx >= 0) 
  {
      laserMsg.ranges[3] = tempProx+ROBOT_RADIUS; 
  } 
  else 
  { 
      laserMsg.ranges[3] = laserMsg.range_max;
  }        
  
  tempProx = proximity_msgs[0].range;
  if(tempProx >= 0) 
  {
      laserMsg.ranges[4] = tempProx+ROBOT_RADIUS; 
  } 
  else 
  { 
      laserMsg.ranges[4] = laserMsg.range_max;
  }        
  
  tempProx = proximity_msgs[1].range;
  if(tempProx >= 0) 
  {
      laserMsg.ranges[5] = tempProx+ROBOT_RADIUS; 
  } 
  else 
  { 
      laserMsg.ranges[5] = laserMsg.range_max;
  }        
  
  tempProx = proximity_msgs[2].range;
  if(tempProx >= 0) 
  {
      laserMsg.ranges[6] = tempProx+ROBOT_RADIUS; 
  } 
  else 
  { 
      laserMsg.ranges[6] = laserMsg.range_max;
  }        
  
  tempProx = proximity_msgs[3].range;
  if(tempProx >= 0) 
  {
      laserMsg.ranges[7] = tempProx+ROBOT_RADIUS; 
  } 
  else 
  { 
      laserMsg.ranges[7] = laserMsg.range_max;
  }                     
  
  for(int i=0; i<8; i++) 
  {
    if(laserMsg.ranges[i] > laserMsg.range_max) 
    {
        laserMsg.ranges[i] = laserMsg.range_max;
    }
    if(laserMsg.ranges[i] < laserMsg.range_min) 
    {
        laserMsg.ranges[i] = laserMsg.range_min;
    }
  }
  
  //laser scan transform
  transform.setOrigin( tf::Vector3(0.0, 0.0, 0.034) );
  q.setRPY(0, 0, 0);
  transform.setRotation(q);
  parent.str("");
  child.str("");
  parent << name.str() << "/base_laser";
  child << name.str() << "/base_link";
  br.sendTransform(tf::StampedTransform(transform, ros::Time::now(), child.str(), parent.str()));
  laserMsg.header.stamp = ros::Time::now();
  laserPublisher.publish(laserMsg);


  //odometry computation
  leftStepsDiff = speedLeft*MOT_STEP_DIST*calibStep ; 
  rightStepsDiff = speedRight*MOT_STEP_DIST*calibStep ; 

  deltaTheta = ((rightStepsDiff - leftStepsDiff)/WHEEL_DISTANCE)*M_PI/calibAngle;   // Expressed in radiant.
  deltaSteps = (rightStepsDiff + leftStepsDiff)/2;        // Expressed in meters.

  xPos += deltaSteps*cos(theta + deltaTheta/2);   // Expressed in meters.
  
  yPos += deltaSteps*sin(theta + deltaTheta/2);   // Expressed in meters.
  
  theta += deltaTheta; // Expressed in radiant.
  if (theta >= 2*M_PI)
  {
    theta -= 2*M_PI;
  }  
  else if (theta <= -2*M_PI)
  {
    theta += 2*M_PI;
  }   
 
  //odometry message 
  odomMsg.header.stamp = ros::Time::now();
  std::stringstream ss;
  ss.str("");
  ss << name.str() << "/odom";
  odomMsg.header.frame_id = ss.str();
  ss.str("");
  ss << name.str() << "/base_link";
  odomMsg.child_frame_id = ss.str();
  odomMsg.pose.pose.position.x = xPos;
  odomMsg.pose.pose.position.y = yPos;
  odomMsg.pose.pose.position.z = 0;
  geometry_msgs::Quaternion odomQuat = tf::createQuaternionMsgFromYaw(theta); // Since all odometry is 6DOF we'll need a quaternion created from yaw.
  odomMsg.pose.pose.orientation = odomQuat;
  currentTime = ros::Time::now();
  odomMsg.twist.twist.linear.x = deltaSteps/ (abs((currentTime-lastTime).toSec()));
  odomMsg.twist.twist.angular.z = deltaTheta/ (abs((currentTime-lastTime).toSec()));
  lastTime = ros::Time::now();
  odomPublisher.publish(odomMsg);

  //odometry transform
  geometry_msgs::TransformStamped odomTrans;
  odomTrans.header.stamp = odomMsg.header.stamp;
  odomTrans.header.frame_id = odomMsg.header.frame_id;
  odomTrans.child_frame_id = odomMsg.child_frame_id;
  odomTrans.transform.translation.x = xPos;
  odomTrans.transform.translation.y = yPos;
  odomTrans.transform.translation.z = 0.0;
  odomTrans.transform.rotation = odomQuat;
  br.sendTransform(odomTrans);

   
}

CVector2 CEPuckRos::ComputeWheelsVelocityFromVector(CVector2 c_vector_to_follow) 
{
  Real fLeftVelocity = 0;
  Real fRightVelocity = 0;
  CRange<CRadians> cLeftHemisphere(CRadians::ZERO, CRadians::PI);
  CRange<CRadians> cRightHemisphere(CRadians::PI, CRadians::TWO_PI);
  CRadians cNormalizedVectorToFollow = c_vector_to_follow.Angle().UnsignedNormalize();

  // Compute relative wheel velocity
  if (c_vector_to_follow.GetX() != 0 || c_vector_to_follow.GetY() != 0) 
  {
    if (cLeftHemisphere.WithinMinBoundExcludedMaxBoundExcluded(cNormalizedVectorToFollow)) 
    {
      fRightVelocity = 1;
      fLeftVelocity = Max<Real>(-0.5f, Cos(cNormalizedVectorToFollow));
    } 
    else 
    {
      fRightVelocity = Max<Real>(-0.5f, Cos(cNormalizedVectorToFollow));
      fLeftVelocity = 1;
    }
  }

  // Transform relative velocity according to max velocity allowed
  Real fVelocityFactor = velocity / Max<Real>(std::abs(fRightVelocity), std::abs(fLeftVelocity));
  CVector2 cWheelsVelocity = CVector2(fVelocityFactor * fLeftVelocity, fVelocityFactor * fRightVelocity);

  return cWheelsVelocity;
}

REGISTER_CONTROLLER(CEPuckRos, "random_walk_controller")
