/*
 * @file <randomWalk.cpp>
 *
 * @author Miquel Kegeleirs - <mkegelei@ulb.ac.be>
 *
 * @license MIT License
 */
#ifndef RANDOMWALK_H
#define RANDOMWALK_H

#include <argos3/core/control_interface/ci_controller.h>
#include <argos3/core/utility/math/vector2.h>
#include <argos3/core/utility/math/rng.h>
#include <argos3/plugins/robots/e-puck/control_interface/ci_epuck_wheels_actuator.h>
#include <argos3/plugins/robots/e-puck/control_interface/ci_epuck_proximity_sensor.h>
#include <argos3/plugins/robots/e-puck/control_interface/ci_epuck_ground_sensor.h>
#include <argos3/plugins/robots/e-puck/control_interface/ci_epuck_battery_sensor.h>
#include <argos3/plugins/robots/e-puck/control_interface/ci_epuck_range_and_bearing_sensor.h>
#include <argos3/plugins/robots/e-puck/control_interface/ci_epuck_range_and_bearing_actuator.h>
#include <argos3/core/utility/configuration/argos_configuration.h>
#include <string>
#include <sstream>
#include <math.h>
#include <vector>

//ros related libraries
#include "ros/ros.h"
#include <tf/transform_broadcaster.h>
#include <sensor_msgs/Range.h>
#include <sensor_msgs/LaserScan.h>
#include <geometry_msgs/Twist.h>
#include <nav_msgs/Odometry.h>

//epuck dimensions parameters
#define ROBOT_RADIUS 0.035
#define WHEEL_DIAMETER 4       
#define WHEEL_SEPARATION 0.53
#define WHEEL_DISTANCE 0.053  
#define WHEEL_CIRCUMFERENCE ((WHEEL_DIAMETER*M_PI)/100.0) 
#define MOT_STEP_DIST (WHEEL_CIRCUMFERENCE/1000.0)

using namespace argos;


class CEPuckRos : public CCI_Controller 
{

public:

   CEPuckRos();

   virtual ~CEPuckRos() {}

   virtual void Init(TConfigurationNode& t_node);

   virtual void ControlStep();

   virtual void Reset() {} 

   virtual void Destroy() {} 

   virtual void initRos(); //must be called in Init

   virtual void rosControlStep(); //must be called in ControlStep

   virtual void walkStep(); //must be called in ControlStep
   
   virtual void vectorFieldAvoid();

   virtual void randomTurnAvoid();

   virtual CVector2 ComputeWheelsVelocityFromVector(CVector2 c_vector_to_follow);

private:

   //actuators and sensors
   CCI_EPuckWheelsActuator* m_pcWheels;
   CCI_EPuckProximitySensor* m_pcProximity;

   //proximity ros publisher
   ros::Publisher proximity_pubs[8];
   sensor_msgs::Range proximity_msgs[8];
   //laser scan ros publisher
   ros::Publisher laserPublisher;
   sensor_msgs::LaserScan laserMsg;
   //odometry ros publisher
   ros::Publisher odomPublisher;
   nav_msgs::Odometry odomMsg;

   //wheels parameters
   Real speedLeft;
   Real speedRight;
   Real velocity;
   //proximity sensors parameters
   double min_range;
   double max_range;
   //odometry parameters
   double leftStepsDiff, rightStepsDiff;
   double leftStepsPrev, rightStepsPrev;
   double xPos, yPos, theta;
   double deltaSteps, deltaTheta;
   double calibAngle, calibStep;
   //random walk parameters
   int randomWalk;
   double mu, gamma;
   int obstacleAvoidance;
   double lMin, bias;
   int stepLength;
   CRadians stepAngle;
   //obstacle avoidance parameters
   int turnCounter;
   bool obstacle;
   double obstacleSensitivity;
   int maxTurnSteps;

   //clocks
   ros::Time currentTime, lastTime;

   //random numbers
   CRandom::CRNG* m_pcRng;
   CRange<Real> m_cRandomRange;

};

#endif
